#include <LiquidCrystal.h>

#define PIN_BUTTON 2
#define PIN_AUTOPLAY 1
#define PIN_READWRITE 10
#define PIN_CONTRAST 12

#define SPRITE_RUN1 1
#define SPRITE_RUN2 2
#define SPRITE_JUMP 3
#define SPRITE_JUMP_UPPER '.'
#define SPRITE_JUMP_LOWER 4
#define SPRITE_GROUND_EMPTY ' '
#define SPRITE_GROUND_SOLID 5
#define SPRITE_GROUND_SOLID_RIGHT 6
#define SPRITE_GROUND_SOLID_LEFT 7

#define GUY_HORIZONTAL_POSITION 1

#define GROUND_WIDTH 16
#define GROUND_EMPTY 0
#define GROUND_LOWER_BLOCK 1
#define GROUND_UPPER_BLOCK 2

#define GUY_POSITION_OFF 0          // Guy is invisible
#define GUY_POSITION_RUN_LOWER_1 1  // Guy is running on lower row (pose 1)
#define GUY_POSITION_RUN_LOWER_2 2  //                              (pose 2)

#define GUY_POSITION_JUMP_1 3       // Starting a jump
#define GUY_POSITION_JUMP_2 4       // Half-way up
#define GUY_POSITION_JUMP_3 5       // Jump is on upper row
#define GUY_POSITION_JUMP_4 6       // Jump is on upper row
#define GUY_POSITION_JUMP_5 7       // Jump is on upper row
#define GUY_POSITION_JUMP_6 8       // Jump is on upper row
#define GUY_POSITION_JUMP_7 9       // Half-way down
#define GUY_POSITION_JUMP_8 10      // About to land

#define GUY_POSITION_RUN_UPPER_1 11 // Guy is running on upper row (pose 1)
#define GUY_POSITION_RUN_UPPER_2 12 //                              (pose 2)

LiquidCrystal lcd(11, 9, 6, 5, 4, 3);
static char groundUpper[GROUND_WIDTH + 1];
static char groundLower[GROUND_WIDTH + 1];
static bool buttonPushed = false;

void initializeGraphics(){
  static byte graphics[] = {
    // Run position 1
    B01100,
    B01100,
    B00000,
    B01110,
    B11100,
    B01100,
    B11010,
    B10011,
    // Run position 2
    B01100,
    B01100,
    B00000,
    B01100,
    B01100,
    B01100,
    B01100,
    B01110,
    // Jump
    B01100,
    B01100,
    B00000,
    B11110,
    B01101,
    B11111,
    B10000,
    B00000,
    // Jump lower
    B11110,
    B01101,
    B11111,
    B10000,
    B00000,
    B00000,
    B00000,
    B00000,
    // Ground
    B11111,
    B11111,
    B11111,
    B11111,
    B11111,
    B11111,
    B11111,
    B11111,
    // Ground right
    B00011,
    B00011,
    B00011,
    B00011,
    B00011,
    B00011,
    B00011,
    B00011,
    // Ground left
    B11000,
    B11000,
    B11000,
    B11000,
    B11000,
    B11000,
    B11000,
    B11000,
  };
  int i;
  for (i = 0; i < 7; ++i) {
	  lcd.createChar(i + 1, &graphics[i * 8]);
  }
  for (i = 0; i < GROUND_WIDTH; ++i) {
    groundUpper[i] = SPRITE_GROUND_EMPTY;
    groundLower[i] = SPRITE_GROUND_EMPTY;
  }
}

void advanceGround(char* ground, byte newGround){
  for (int i = 0; i < GROUND_WIDTH; ++i) {
    char current = ground[i];
    char next = (i == GROUND_WIDTH-1) ? newGround : ground[i+1];
    switch (current){
      case SPRITE_GROUND_EMPTY:
        ground[i] = (next == SPRITE_GROUND_SOLID) ? SPRITE_GROUND_SOLID_RIGHT : SPRITE_GROUND_EMPTY;
        break;
      case SPRITE_GROUND_SOLID:
        ground[i] = (next == SPRITE_GROUND_EMPTY) ? SPRITE_GROUND_SOLID_LEFT : SPRITE_GROUND_SOLID;
        break;
      case SPRITE_GROUND_SOLID_RIGHT:
        ground[i] = SPRITE_GROUND_SOLID;
        break;
      case SPRITE_GROUND_SOLID_LEFT:
        ground[i] = SPRITE_GROUND_EMPTY;
        break;
    }
  }
}

bool drawGuy(byte position, char* groundUpper, char* groundLower, unsigned int score) {
  bool collide = false;
  char upperSave = groundUpper[GUY_HORIZONTAL_POSITION];
  char lowerSave = groundLower[GUY_HORIZONTAL_POSITION];
  byte upper, lower;
  switch (position) {
    case GUY_POSITION_OFF:
      upper = lower = SPRITE_GROUND_EMPTY;
      break;
    case GUY_POSITION_RUN_LOWER_1:
      upper = SPRITE_GROUND_EMPTY;
      lower = SPRITE_RUN1;
      break;
    case GUY_POSITION_RUN_LOWER_2:
      upper = SPRITE_GROUND_EMPTY;
      lower = SPRITE_RUN2;
      break;
    case GUY_POSITION_JUMP_1:
    case GUY_POSITION_JUMP_8:
      upper = SPRITE_GROUND_EMPTY;
      lower = SPRITE_JUMP;
      break;
    case GUY_POSITION_JUMP_2:
    case GUY_POSITION_JUMP_7:
      upper = SPRITE_JUMP_UPPER;
      lower = SPRITE_JUMP_LOWER;
      break;
    case GUY_POSITION_JUMP_3:
    case GUY_POSITION_JUMP_4:
    case GUY_POSITION_JUMP_5:
    case GUY_POSITION_JUMP_6:
      upper = SPRITE_JUMP;
      lower = SPRITE_GROUND_EMPTY;
      break;
    case GUY_POSITION_RUN_UPPER_1:
      upper = SPRITE_RUN1;
      lower = SPRITE_GROUND_EMPTY;
      break;
    case GUY_POSITION_RUN_UPPER_2:
      upper = SPRITE_RUN2;
      lower = SPRITE_GROUND_EMPTY;
      break;
  }
  if (upper != ' ') {
    groundUpper[GUY_HORIZONTAL_POSITION] = upper;
    collide = (upperSave == SPRITE_GROUND_EMPTY) ? false : true;
  }
  if (lower != ' ') {
    groundLower[GUY_HORIZONTAL_POSITION] = lower;
    collide |= (lowerSave == SPRITE_GROUND_EMPTY) ? false : true;
  }
  
  byte digits = (score > 9999) ? 5 : (score > 999) ? 4 : (score > 99) ? 3 : (score > 9) ? 2 : 1;
  
  groundUpper[GROUND_WIDTH] = '\0';
  groundLower[GROUND_WIDTH] = '\0';
  char temp = groundUpper[16-digits];
  groundUpper[16-digits] = '\0';
  lcd.setCursor(0,0);
  lcd.print(groundUpper);
  groundUpper[16-digits] = temp;  
  lcd.setCursor(0,1);
  lcd.print(groundLower);
  
  lcd.setCursor(16 - digits,0);
  lcd.print(score);

  groundUpper[GUY_HORIZONTAL_POSITION] = upperSave;
  groundLower[GUY_HORIZONTAL_POSITION] = lowerSave;
  return collide;
}

void buttonPush() {
  buttonPushed = true;
}

void setup(){
  pinMode(PIN_READWRITE, OUTPUT);
  digitalWrite(PIN_READWRITE, LOW);
  pinMode(PIN_CONTRAST, OUTPUT);
  digitalWrite(PIN_CONTRAST, LOW);
  pinMode(PIN_BUTTON, INPUT);
  digitalWrite(PIN_BUTTON, HIGH);
  pinMode(PIN_AUTOPLAY, OUTPUT);
  digitalWrite(PIN_AUTOPLAY, HIGH);
  
  
  attachInterrupt(0, buttonPush, FALLING);
  
  initializeGraphics();
  
  lcd.begin(16, 2);
}

void loop(){
  static byte guyPos = GUY_POSITION_RUN_LOWER_1;
  static byte newGroundType = GROUND_EMPTY;
  static byte newGroundDuration = 1;
  static bool playing = false;
  static bool blink = false;
  static unsigned int distance = 0;
  
  if (!playing) {
    drawGuy((blink) ? GUY_POSITION_OFF : guyPos, groundUpper, groundLower, distance >> 3);
    if (blink) {
      lcd.setCursor(0,0);
      lcd.print("Press Start");
    }
    delay(250);
    blink = !blink;
    if (buttonPushed) {
      initializeGraphics();
      guyPos = GUY_POSITION_RUN_LOWER_1;
      playing = true;
      buttonPushed = false;
      distance = 0;
    }
    return;
  }

  advanceGround(groundLower, newGroundType == GROUND_LOWER_BLOCK ? SPRITE_GROUND_SOLID : SPRITE_GROUND_EMPTY);
  advanceGround(groundUpper, newGroundType == GROUND_UPPER_BLOCK ? SPRITE_GROUND_SOLID : SPRITE_GROUND_EMPTY);
  
  if (--newGroundDuration == 0) {
    if (newGroundType == GROUND_EMPTY) {
      newGroundType = (random(3) == 0) ? GROUND_UPPER_BLOCK : GROUND_LOWER_BLOCK;
      newGroundDuration = 2 + random(10);
    } else {
      newGroundType = GROUND_EMPTY;
      newGroundDuration = 10 + random(10);
    }
  }
    
  if (buttonPushed) {
    if (guyPos <= GUY_POSITION_RUN_LOWER_2) guyPos = GUY_POSITION_JUMP_1;
    buttonPushed = false;
  }  

  if (drawGuy(guyPos, groundUpper, groundLower, distance >> 3)) {
    playing = false; // Guy collided with something.
  } else {
    if (guyPos == GUY_POSITION_RUN_LOWER_2 || guyPos == GUY_POSITION_JUMP_8) {
      guyPos = GUY_POSITION_RUN_LOWER_1;
    } else if ((guyPos >= GUY_POSITION_JUMP_3 && guyPos <= GUY_POSITION_JUMP_5) && groundLower[GUY_HORIZONTAL_POSITION] != SPRITE_GROUND_EMPTY) {
      guyPos = GUY_POSITION_RUN_UPPER_1;
    } else if (guyPos >= GUY_POSITION_RUN_UPPER_1 && groundLower[GUY_HORIZONTAL_POSITION] == SPRITE_GROUND_EMPTY) {
      guyPos = GUY_POSITION_JUMP_5;
    } else if (guyPos == GUY_POSITION_RUN_UPPER_2) {
      guyPos = GUY_POSITION_RUN_UPPER_1;
    } else {
      ++guyPos;
    }
    ++distance;
    
    digitalWrite(PIN_AUTOPLAY, groundLower[GUY_HORIZONTAL_POSITION + 2] == SPRITE_GROUND_EMPTY ? HIGH : LOW);
  }
  delay(50);
}
